Duration

Struct Duration 

pub struct Duration { /* private fields */ }
Defines generally usable durations for nanosecond precision valid for 32,768 centuries in either direction, and only on 80 bits / 10 octets.

Important conventions:

  1. The negative durations can be mentally modeled “BC” years. One hours before 01 Jan 0000, it was “-1” years but 365 days and 23h into the current day. It was decided that the nanoseconds corresponds to the nanoseconds into the current century. In other words, a duration with centuries = -1 and nanoseconds = 0 is a greater duration (further from zero) than centuries = -1 and nanoseconds = 1. Duration zero minus one nanosecond returns a century of -1 and a nanosecond set to the number of nanoseconds in one century minus one. That difference is exactly 1 nanoseconds, where the former duration is “closer to zero” than the latter. As such, the largest negative duration that can be represented sets the centuries to i16::MAX and its nanoseconds to NANOSECONDS_PER_CENTURY.
  2. Negative and positive durations are distinct: -15 minutes != 15 minutes. Use the signum function to check the sign, and abs() to get the absolute value.

:type string_repr: str

Implementations



impl Duration

pub const fn py_to_parts(&self) -> (i16, u64)

Returns the centuries and nanoseconds of this duration NOTE: These items are not public to prevent incorrect durations from being created by modifying the values of the structure directly. :rtype: tuple[int, int]

pub fn py_total_nanoseconds(&self) -> i128

Returns the total nanoseconds in a signed 128 bit integer :rtype: int

pub fn py_to_seconds(&self) -> f64

Returns this duration in seconds f64. For high fidelity comparisons, it is recommended to keep using the Duration structure. :rtype: float

pub fn py_to_unit(&self, unit: Unit) -> f64

:type unit: Unit :rtype: float

pub fn py_abs(&self) -> Duration

Returns the absolute value of this duration :rtype: Duration

pub const fn py_signum(&self) -> i8

Returns the sign of this duration

  • 0 if the number is zero
  • 1 if the number is positive
  • -1 if the number is negative :rtype: int

pub fn py_decompose(&self) -> (i8, u64, u64, u64, u64, u64, u64, u64)

Decomposes a Duration in its sign, days, hours, minutes, seconds, ms, us, ns

:rtype: tuple[int, int, int, int, int, int, int, int]

pub fn py_floor(&self, duration: Duration) -> Duration

Floors this duration to the closest duration from the bottom

Example
use hifitime::{Duration, TimeUnits};

let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.floor(1.hours()), 2.hours());
assert_eq!(two_hours_three_min.floor(30.minutes()), 2.hours());
// This is zero because we floor by a duration longer than the current duration, rounding it down
assert_eq!(two_hours_three_min.floor(4.hours()), 0.hours());
assert_eq!(two_hours_three_min.floor(1.seconds()), two_hours_three_min);
assert_eq!(two_hours_three_min.floor(1.hours() + 1.minutes()), 2.hours() + 2.minutes());
assert_eq!(two_hours_three_min.floor(1.hours() + 5.minutes()), 1.hours() + 5.minutes());

:type duration: Duration :rtype: Duration

pub fn py_ceil(&self, duration: Duration) -> Duration

Ceils this duration to the closest provided duration

This simply floors then adds the requested duration

Example
use hifitime::{Duration, TimeUnits};

let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.ceil(1.hours()), 3.hours());
assert_eq!(two_hours_three_min.ceil(30.minutes()), 2.hours() + 30.minutes());
assert_eq!(two_hours_three_min.ceil(4.hours()), 4.hours());
assert_eq!(two_hours_three_min.ceil(1.seconds()), two_hours_three_min + 1.seconds());
assert_eq!(two_hours_three_min.ceil(1.hours() + 5.minutes()), 2.hours() + 10.minutes());

:type duration: Duration :rtype: Duration

pub fn py_round(&self, duration: Duration) -> Duration

Rounds this duration to the closest provided duration

This performs both a ceil and floor and returns the value which is the closest to current one.

Example
use hifitime::{Duration, TimeUnits};

let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.round(1.hours()), 2.hours());
assert_eq!(two_hours_three_min.round(30.minutes()), 2.hours());
assert_eq!(two_hours_three_min.round(4.hours()), 4.hours());
assert_eq!(two_hours_three_min.round(1.seconds()), two_hours_three_min);
assert_eq!(two_hours_three_min.round(1.hours() + 5.minutes()), 2.hours() + 10.minutes());

:type duration: Duration :rtype: Duration

pub fn py_approx(&self) -> Duration

Rounds this duration to the largest units represented in this duration.

This is useful to provide an approximate human duration. Under the hood, this function uses round, so the “tipping point” of the rounding is half way to the next increment of the greatest unit. As shown below, one example is that 35 hours and 59 minutes rounds to 1 day, but 36 hours and 1 minute rounds to 2 days because 2 days is closer to 36h 1 min than 36h 1 min is to 1 day.

Example
use hifitime::{Duration, TimeUnits};

assert_eq!((2.hours() + 3.minutes()).approx(), 2.hours());
assert_eq!((24.hours() + 3.minutes()).approx(), 1.days());
assert_eq!((35.hours() + 59.minutes()).approx(), 1.days());
assert_eq!((36.hours() + 1.minutes()).approx(), 2.days());
assert_eq!((47.hours() + 3.minutes()).approx(), 2.days());
assert_eq!((49.hours() + 3.minutes()).approx(), 2.days());

:rtype: Duration

pub fn py_min(&self, other: Duration) -> Duration

Returns the minimum of the two durations.

use hifitime::TimeUnits;

let d0 = 20.seconds();
let d1 = 21.seconds();

assert_eq!(d0, d1.min(d0));
assert_eq!(d0, d0.min(d1));

:type other: Duration :rtype: Duration

pub fn py_max(&self, other: Duration) -> Duration

Returns the maximum of the two durations.

use hifitime::TimeUnits;

let d0 = 20.seconds();
let d1 = 21.seconds();

assert_eq!(d1, d1.max(d0));
assert_eq!(d1, d0.max(d1));

:type other: Duration :rtype: Duration

pub fn py_is_negative(&self) -> bool

Returns whether this is a negative or positive duration. :rtype: bool



impl Duration

pub const ZERO: Duration

A duration of exactly zero nanoseconds

pub const MAX: Duration

Maximum duration that can be represented

pub const MIN: Duration

Minimum duration that can be represented

pub const EPSILON: Duration

Smallest duration that can be represented

pub const MIN_POSITIVE: Duration = Self::EPSILON

Minimum positive duration is one nanoseconds

pub const MIN_NEGATIVE: Duration

Minimum negative duration is minus one nanosecond

pub const fn from_parts(centuries: i16, nanoseconds: u64) -> Duration

Create a normalized duration from its parts

pub const fn from_total_nanoseconds(nanos: i128) -> Duration

Converts the total nanoseconds as i128 into this Duration (saving 48 bits)

pub const fn from_truncated_nanoseconds(nanos: i64) -> Duration

Create a new duration from the truncated nanoseconds (+/- 2927.1 years of duration)

pub const fn from_days(value: f64) -> Duration

Creates a new duration from the provided number of days

pub const fn from_hours(value: f64) -> Duration

Creates a new duration from the provided number of hours

pub const fn from_seconds(value: f64) -> Duration

Creates a new duration from the provided number of seconds

pub const fn from_milliseconds(value: f64) -> Duration

Creates a new duration from the provided number of milliseconds

pub const fn from_microseconds(value: f64) -> Duration

Creates a new duration from the provided number of microsecond

pub const fn from_nanoseconds(value: f64) -> Duration

Creates a new duration from the provided number of nanoseconds

pub fn compose( sign: i8, days: u64, hours: u64, minutes: u64, seconds: u64, milliseconds: u64, microseconds: u64, nanoseconds: u64, ) -> Duration

Creates a new duration from its parts. Set the sign to a negative number for the duration to be negative.

pub fn compose_f64( sign: i8, days: f64, hours: f64, minutes: f64, seconds: f64, milliseconds: f64, microseconds: f64, nanoseconds: f64, ) -> Duration

Creates a new duration from its parts. Set the sign to a negative number for the duration to be negative.

pub fn from_tz_offset(sign: i8, hours: i64, minutes: i64) -> Duration

Initializes a Duration from a timezone offset



impl Duration

pub const fn to_parts(&self) -> (i16, u64)

Returns the centuries and nanoseconds of this duration NOTE: These items are not public to prevent incorrect durations from being created by modifying the values of the structure directly.

pub fn total_nanoseconds(&self) -> i128

Returns the total nanoseconds in a signed 128 bit integer

pub fn try_truncated_nanoseconds(&self) -> Result<i64, HifitimeError>

Returns the truncated nanoseconds in a signed 64 bit integer, if the duration fits.

pub fn truncated_nanoseconds(&self) -> i64

Returns the truncated nanoseconds in a signed 64 bit integer, if the duration fits. WARNING: This function will NOT fail and will return the i64::MIN or i64::MAX depending on the sign of the centuries if the Duration does not fit on aa i64

pub fn to_seconds(&self) -> f64

Returns this duration in seconds f64. For high fidelity comparisons, it is recommended to keep using the Duration structure.

Examples found in repository?
nyx-core/examples/03_geo_analysis/drift.rs (line 178)
26fn main() -> Result<(), Box<dyn Error>> {
27    pel::init();
28    // Dynamics models require planetary constants and ephemerides to be defined.
29    // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
30    // This will automatically download the DE440s planetary ephemeris,
31    // the daily-updated Earth Orientation Parameters, the high fidelity Moon orientation
32    // parameters (for the Moon Mean Earth and Moon Principal Axes frames), and the PCK11
33    // planetary constants kernels.
34    // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
35    // Note that we place the Almanac into an Arc so we can clone it cheaply and provide read-only
36    // references to many functions.
37    let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
38    // Define the orbit epoch
39    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
40
41    // Define the orbit.
42    // First we need to fetch the Earth J2000 from information from the Almanac.
43    // This allows the frame to include the gravitational parameters and the shape of the Earth,
44    // defined as a tri-axial ellipoid. Note that this shape can be changed manually or in the Almanac
45    // by loading a different set of planetary constants.
46    let earth_j2000 = almanac.frame_info(EARTH_J2000)?;
47
48    // Placing this GEO bird just above Colorado.
49    // In theory, the eccentricity is zero, but in practice, it's about 1e-5 to 1e-6 at best.
50    let orbit = Orbit::try_keplerian(42164.0, 1e-5, 0., 163.0, 75.0, 0.0, epoch, earth_j2000)?;
51    // Print in in Keplerian form.
52    println!("{orbit:x}");
53
54    let state_bf = almanac.transform_to(orbit, IAU_EARTH_FRAME, None)?;
55    let (orig_lat_deg, orig_long_deg, orig_alt_km) = state_bf.latlongalt()?;
56
57    // Nyx is used for high fidelity propagation, not Keplerian propagation as above.
58    // Nyx only propagates Spacecraft at the moment, which allows it to account for acceleration
59    // models such as solar radiation pressure.
60
61    // Let's build a cubesat sized spacecraft, with an SRP area of 10 cm^2 and a mass of 9.6 kg.
62    let sc = Spacecraft::builder()
63        .orbit(orbit)
64        .mass(Mass::from_dry_mass(9.60))
65        .srp(SRPData {
66            area_m2: 10e-4,
67            coeff_reflectivity: 1.1,
68        })
69        .build();
70    println!("{sc:x}");
71
72    // Set up the spacecraft dynamics.
73
74    // Specify that the orbital dynamics must account for the graviational pull of the Moon and the Sun.
75    // The gravity of the Earth will also be accounted for since the spaceraft in an Earth orbit.
76    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
77
78    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
79    // We're using the JGM3 model here, which is the default in GMAT.
80    let mut jgm3_meta = MetaFile {
81        uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
82        crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
83    };
84    // And let's download it if we don't have it yet.
85    jgm3_meta.process(true)?;
86
87    // Build the spherical harmonics.
88    // The harmonics must be computed in the body fixed frame.
89    // We're using the long term prediction of the Earth centered Earth fixed frame, IAU Earth.
90    let harmonics_21x21 = GravityField::new(
91        GravityFieldData::from_cof(
92            &jgm3_meta.uri,
93            21,
94            21,
95            true,
96            almanac.frame_info(IAU_EARTH_FRAME)?,
97        )
98        .unwrap(),
99    );
100
101    // Include the spherical harmonics into the orbital dynamics.
102    orbital_dyn.accel_models.push(harmonics_21x21);
103
104    // We define the solar radiation pressure, using the default solar flux and accounting only
105    // for the eclipsing caused by the Earth and Moon.
106    let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], &almanac)?;
107
108    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
109    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
110    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
111
112    println!("{dynamics}");
113
114    // Finally, let's propagate this orbit to the same epoch as above.
115    // The first returned value is the spacecraft state at the final epoch.
116    // The second value is the full trajectory where the step size is variable step used by the propagator.
117    let (future_sc, trajectory) = Propagator::default(dynamics)
118        .with(sc, almanac.clone())
119        .until_epoch_with_traj(epoch + Unit::Century * 0.03)?;
120
121    println!("=== High fidelity propagation ===");
122    println!(
123        "SMA changed by {:.3} km",
124        orbit.sma_km()? - future_sc.orbit.sma_km()?
125    );
126    println!(
127        "ECC changed by {:.6}",
128        orbit.ecc()? - future_sc.orbit.ecc()?
129    );
130    println!(
131        "INC changed by {:.3e} deg",
132        orbit.inc_deg()? - future_sc.orbit.inc_deg()?
133    );
134    println!(
135        "RAAN changed by {:.3} deg",
136        orbit.raan_deg()? - future_sc.orbit.raan_deg()?
137    );
138    println!(
139        "AOP changed by {:.3} deg",
140        orbit.aop_deg()? - future_sc.orbit.aop_deg()?
141    );
142    println!(
143        "TA changed by {:.3} deg",
144        orbit.ta_deg()? - future_sc.orbit.ta_deg()?
145    );
146
147    // We also have access to the full trajectory throughout the propagation.
148    println!("{trajectory}");
149
150    println!("Spacecraft params after 3 years without active control:\n{future_sc:x}");
151
152    // With the trajectory, let's build a few data products.
153
154    // 1. Export the trajectory as a parquet file, which includes the Keplerian orbital elements.
155
156    let analysis_step = Unit::Minute * 5;
157
158    trajectory.to_parquet(
159        "./03_geo_hf_prop.parquet",
160        ExportCfg::builder().step(analysis_step).build(),
161    )?;
162
163    // 2. Compute the latitude, longitude, and altitude throughout the trajectory by rotating the spacecraft position into the Earth body fixed frame.
164
165    // We iterate over the trajectory, grabbing a state every two minutes.
166    let mut offset_s = vec![];
167    let mut epoch_str = vec![];
168    let mut longitude_deg = vec![];
169    let mut latitude_deg = vec![];
170    let mut altitude_km = vec![];
171
172    for state in trajectory.every(analysis_step) {
173        // Convert the GEO bird state into the body fixed frame, and keep track of its latitude, longitude, and altitude.
174        // These define the GEO stationkeeping box.
175
176        let this_epoch = state.epoch();
177
178        offset_s.push((this_epoch - orbit.epoch).to_seconds());
179        epoch_str.push(this_epoch.to_isoformat());
180
181        let state_bf = almanac.transform_to(state.orbit, IAU_EARTH_FRAME, None)?;
182        let (lat_deg, long_deg, alt_km) = state_bf.latlongalt()?;
183        longitude_deg.push(long_deg);
184        latitude_deg.push(lat_deg);
185        altitude_km.push(alt_km);
186    }
187
188    println!(
189        "Longitude changed by {:.3} deg -- Box is 0.1 deg E-W",
190        orig_long_deg - longitude_deg.last().unwrap()
191    );
192
193    println!(
194        "Latitude changed by {:.3} deg -- Box is 0.05 deg N-S",
195        orig_lat_deg - latitude_deg.last().unwrap()
196    );
197
198    println!(
199        "Altitude changed by {:.3} km -- Box is 30 km",
200        orig_alt_km - altitude_km.last().unwrap()
201    );
202
203    // Build the station keeping data frame.
204    let mut sk_df = df!(
205        "Offset (s)" => offset_s.clone(),
206        "Epoch (UTC)" => epoch_str.clone(),
207        "Longitude E-W (deg)" => longitude_deg,
208        "Latitude N-S (deg)" => latitude_deg,
209        "Altitude (km)" => altitude_km,
210
211    )?;
212
213    // Create a file to write the Parquet to
214    let file = File::create("./03_geo_lla.parquet").expect("Could not create file");
215
216    // Create a ParquetWriter and write the DataFrame to the file
217    ParquetWriter::new(file).finish(&mut sk_df)?;
218
219    Ok(())
220}
More examples
Hide additional examples
nyx-core/examples/01_orbit_prop/main.rs (line 236)
30fn main() -> Result<(), Box<dyn Error>> {
31    pel::init();
32    // Dynamics models require planetary constants and ephemerides to be defined.
33    // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
34    // This will automatically download the DE440s planetary ephemeris,
35    // the daily-updated Earth Orientation Parameters, the high fidelity Moon orientation
36    // parameters (for the Moon Mean Earth and Moon Principal Axes frames), and the PCK11
37    // planetary constants kernels.
38    // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
39    // Note that we place the Almanac into an Arc so we can clone it cheaply and provide read-only
40    // references to many functions.
41    let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
42    // Define the orbit epoch
43    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
44
45    // Define the orbit.
46    // First we need to fetch the Earth J2000 from information from the Almanac.
47    // This allows the frame to include the gravitational parameters and the shape of the Earth,
48    // defined as a tri-axial ellipoid. Note that this shape can be changed manually or in the Almanac
49    // by loading a different set of planetary constants.
50    let earth_j2000 = almanac.frame_info(EARTH_J2000)?;
51
52    let orbit =
53        Orbit::try_keplerian_altitude(300.0, 0.015, 68.5, 65.2, 75.0, 0.0, epoch, earth_j2000)?;
54    // Print in in Keplerian form.
55    println!("{orbit:x}");
56
57    // There are two ways to propagate an orbit. We can make a quick approximation assuming only two-body
58    // motion. This is a useful first order approximation but it isn't used in real-world applications.
59
60    // This approach is a feature of ANISE.
61    let future_orbit_tb = orbit.at_epoch(epoch + Unit::Day * 3)?;
62    println!("{future_orbit_tb:x}");
63
64    // Two body propagation relies solely on Kepler's laws, so only the true anomaly will change.
65    println!(
66        "SMA changed by {:.3e} km",
67        orbit.sma_km()? - future_orbit_tb.sma_km()?
68    );
69    println!(
70        "ECC changed by {:.3e}",
71        orbit.ecc()? - future_orbit_tb.ecc()?
72    );
73    println!(
74        "INC changed by {:.3e} deg",
75        orbit.inc_deg()? - future_orbit_tb.inc_deg()?
76    );
77    println!(
78        "RAAN changed by {:.3e} deg",
79        orbit.raan_deg()? - future_orbit_tb.raan_deg()?
80    );
81    println!(
82        "AOP changed by {:.3e} deg",
83        orbit.aop_deg()? - future_orbit_tb.aop_deg()?
84    );
85    println!(
86        "TA changed by {:.3} deg",
87        orbit.ta_deg()? - future_orbit_tb.ta_deg()?
88    );
89
90    // Nyx is used for high fidelity propagation, not Keplerian propagation as above.
91    // Nyx only propagates Spacecraft at the moment, which allows it to account for acceleration
92    // models such as solar radiation pressure.
93
94    // Let's build a cubesat sized spacecraft, with an SRP area of 10 cm^2 and a mass of 9.6 kg.
95    let sc = Spacecraft::builder()
96        .orbit(orbit)
97        .mass(Mass::from_dry_mass(9.60))
98        .srp(SRPData {
99            area_m2: 10e-4,
100            coeff_reflectivity: 1.1,
101        })
102        .build();
103    println!("{sc:x}");
104
105    // Set up the spacecraft dynamics.
106
107    // Specify that the orbital dynamics must account for the graviational pull of the Moon and the Sun.
108    // The gravity of the Earth will also be accounted for since the spaceraft in an Earth orbit.
109    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
110
111    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
112    // We're using the JGM3 model here, which is the default in GMAT.
113    let mut jgm3_meta = MetaFile {
114        uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
115        crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
116    };
117    // And let's download it if we don't have it yet.
118    jgm3_meta.process(true)?;
119
120    // Build the spherical harmonics.
121    // The harmonics must be computed in the body fixed frame.
122    // We're using the long term prediction of the Earth centered Earth fixed frame, IAU Earth.
123    let harmonics_21x21 = GravityField::new(
124        GravityFieldData::from_cof(
125            &jgm3_meta.uri,
126            21,
127            21,
128            true,
129            almanac.frame_info(IAU_EARTH_FRAME)?,
130        )
131        .unwrap(),
132    );
133
134    // Include the spherical harmonics into the orbital dynamics.
135    orbital_dyn.accel_models.push(harmonics_21x21);
136
137    // We define the solar radiation pressure, using the default solar flux and accounting only
138    // for the eclipsing caused by the Earth.
139    let srp_dyn = SolarPressure::default_flux(EARTH_J2000, &almanac)?;
140
141    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
142    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
143    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
144
145    println!("{dynamics}");
146
147    // Finally, let's propagate this orbit to the same epoch as above.
148    // The first returned value is the spacecraft state at the final epoch.
149    // The second value is the full trajectory where the step size is variable step used by the propagator.
150    let (future_sc, trajectory) = Propagator::default(dynamics)
151        .with(sc, almanac.clone())
152        .until_epoch_with_traj(future_orbit_tb.epoch)?;
153
154    println!("=== High fidelity propagation ===");
155    println!(
156        "SMA changed by {:.3} km",
157        orbit.sma_km()? - future_sc.orbit.sma_km()?
158    );
159    println!(
160        "ECC changed by {:.6}",
161        orbit.ecc()? - future_sc.orbit.ecc()?
162    );
163    println!(
164        "INC changed by {:.3e} deg",
165        orbit.inc_deg()? - future_sc.orbit.inc_deg()?
166    );
167    println!(
168        "RAAN changed by {:.3} deg",
169        orbit.raan_deg()? - future_sc.orbit.raan_deg()?
170    );
171    println!(
172        "AOP changed by {:.3} deg",
173        orbit.aop_deg()? - future_sc.orbit.aop_deg()?
174    );
175    println!(
176        "TA changed by {:.3} deg",
177        orbit.ta_deg()? - future_sc.orbit.ta_deg()?
178    );
179
180    // We also have access to the full trajectory throughout the propagation.
181    println!("{trajectory}");
182
183    // With the trajectory, let's build a few data products.
184
185    // 1. Export the trajectory as a CCSDS OEM version 2.0 file and as a parquet file, which includes the Keplerian orbital elements.
186
187    trajectory.to_oem_file(
188        "./01_cubesat_hf_prop.oem",
189        "CUBESAT-ID".to_string(),
190        Some("Nyx Space".to_string()),
191        Some("CUBESAT".to_string()),
192        ExportCfg::builder().step(Unit::Minute * 2).build(),
193    )?;
194
195    trajectory.to_parquet_with_cfg(
196        "./01_cubesat_hf_prop.parquet",
197        ExportCfg::builder().step(Unit::Minute * 2).build(),
198    )?;
199
200    // 2. Compare the difference in the radial-intrack-crosstrack frame between the high fidelity
201    // and Keplerian propagation. The RIC frame is commonly used to compute the difference in position
202    // and velocity of different spacecraft.
203    // 3. Compute the azimuth, elevation, range, and range-rate data of that spacecraft as seen from Boulder, CO, USA.
204
205    let boulder_station = GroundStation::from_point(
206        "Boulder, CO, USA".to_string(),
207        40.014984,   // latitude in degrees
208        -105.270546, // longitude in degrees
209        1.6550,      // altitude in kilometers
210        almanac.frame_info(IAU_EARTH_FRAME)?,
211    );
212
213    // We iterate over the trajectory, grabbing a state every two minutes.
214    let mut offset_s = vec![];
215    let mut epoch_str = vec![];
216    let mut ric_x_km = vec![];
217    let mut ric_y_km = vec![];
218    let mut ric_z_km = vec![];
219    let mut ric_vx_km_s = vec![];
220    let mut ric_vy_km_s = vec![];
221    let mut ric_vz_km_s = vec![];
222
223    let mut azimuth_deg = vec![];
224    let mut elevation_deg = vec![];
225    let mut range_km = vec![];
226    let mut range_rate_km_s = vec![];
227    for state in trajectory.every(Unit::Minute * 2) {
228        // Try to compute the Keplerian/two body state just in time.
229        // This method occasionally fails to converge on an appropriate true anomaly
230        // from the mean anomaly. If that happens, we just skip this state.
231        // The high fidelity and Keplerian states diverge continuously, and we're curious
232        // about the divergence in this quick analysis.
233        let this_epoch = state.epoch();
234        match orbit.at_epoch(this_epoch) {
235            Ok(tb_then) => {
236                offset_s.push((this_epoch - orbit.epoch).to_seconds());
237                epoch_str.push(format!("{this_epoch}"));
238                // Compute the two body state just in time.
239                let ric = state.orbit.ric_difference(&tb_then)?;
240                ric_x_km.push(ric.radius_km.x);
241                ric_y_km.push(ric.radius_km.y);
242                ric_z_km.push(ric.radius_km.z);
243                ric_vx_km_s.push(ric.velocity_km_s.x);
244                ric_vy_km_s.push(ric.velocity_km_s.y);
245                ric_vz_km_s.push(ric.velocity_km_s.z);
246
247                // Compute the AER data for each state.
248                let aer = almanac.azimuth_elevation_range_sez(
249                    state.orbit,
250                    boulder_station.to_orbit(this_epoch, &almanac)?,
251                    None,
252                    None,
253                )?;
254                azimuth_deg.push(aer.azimuth_deg);
255                elevation_deg.push(aer.elevation_deg);
256                range_km.push(aer.range_km);
257                range_rate_km_s.push(aer.range_rate_km_s);
258            }
259            Err(e) => warn!("{} {e}", state.epoch()),
260        };
261    }
262
263    // Build the data frames.
264    let ric_df = df!(
265        "Offset (s)" => offset_s.clone(),
266        "Epoch" => epoch_str.clone(),
267        "RIC X (km)" => ric_x_km,
268        "RIC Y (km)" => ric_y_km,
269        "RIC Z (km)" => ric_z_km,
270        "RIC VX (km/s)" => ric_vx_km_s,
271        "RIC VY (km/s)" => ric_vy_km_s,
272        "RIC VZ (km/s)" => ric_vz_km_s,
273    )?;
274
275    println!("RIC difference at start\n{}", ric_df.head(Some(10)));
276    println!("RIC difference at end\n{}", ric_df.tail(Some(10)));
277
278    let aer_df = df!(
279        "Offset (s)" => offset_s.clone(),
280        "Epoch" => epoch_str.clone(),
281        "azimuth (deg)" => azimuth_deg,
282        "elevation (deg)" => elevation_deg,
283        "range (km)" => range_km,
284        "range rate (km/s)" => range_rate_km_s,
285    )?;
286
287    // Finally, let's see when the spacecraft is visible, assuming 15 degrees minimum elevation.
288    let mask = aer_df
289        .column("elevation (deg)")?
290        .gt(&Column::Scalar(ScalarColumn::new(
291            "elevation mask (deg)".into(),
292            Scalar::new(DataType::Float64, AnyValue::Float64(15.0)),
293            offset_s.len(),
294        )))?;
295    let cubesat_visible = aer_df.filter(&mask)?;
296
297    println!("{cubesat_visible}");
298
299    Ok(())
300}

pub fn to_unit(&self, unit: Unit) -> f64

pub fn abs(&self) -> Duration

Returns the absolute value of this duration

Examples found in repository?
nyx-core/examples/04_lro_od/main.rs (line 355)
35fn main() -> Result<(), Box<dyn Error>> {
36    pel::init();
37
38    // ====================== //
39    // === ALMANAC SET UP === //
40    // ====================== //
41
42    // Dynamics models require planetary constants and ephemerides to be defined.
43    // Let's start by grabbing those by using ANISE's MetaAlmanac.
44
45    let output_folder: PathBuf = [env!("CARGO_MANIFEST_DIR"), "../data", "04_output"]
46        .iter()
47        .collect();
48
49    let data_folder: PathBuf = [env!("CARGO_MANIFEST_DIR"), "examples", "04_lro_od"]
50        .iter()
51        .collect();
52
53    let meta = data_folder.join("lro-dynamics.dhall");
54
55    // Load this ephem in the general Almanac we're using for this analysis.
56    let mut almanac = MetaAlmanac::new(meta.to_string_lossy().as_ref())
57        .map_err(Box::new)?
58        .process(true)
59        .map_err(Box::new)?;
60
61    let mut moon_pc = almanac.get_planetary_data_from_id(MOON).unwrap();
62    moon_pc.mu_km3_s2 = 4902.74987;
63    almanac.set_planetary_data_from_id(MOON, moon_pc).unwrap();
64
65    let mut earth = almanac.get_planetary_data_from_id(EARTH).unwrap();
66    earth.mu_km3_s2 = 398600.436;
67    almanac.set_planetary_data_from_id(EARTH, earth).unwrap();
68
69    // Save this new kernel for reuse.
70    // In an operational context, this would be part of the "Lock" process, and should not change throughout the mission.
71    almanac
72        .planetary_data
73        .values()
74        .next()
75        .unwrap()
76        .save_as(&data_folder.join("lro-specific.pca"), true)?;
77
78    // Lock the almanac (an Arc is a read only structure).
79    let almanac = Arc::new(almanac);
80
81    // Orbit determination requires a Trajectory structure, which can be saved as parquet file.
82    // In our case, the trajectory comes from the BSP file, so we need to build a Trajectory from the almanac directly.
83    // To query the Almanac, we need to build the LRO frame in the J2000 orientation in our case.
84    // Inspecting the LRO BSP in the ANISE GUI shows us that NASA has assigned ID -85 to LRO.
85    let lro_frame = Frame::from_ephem_j2000(-85);
86
87    // To build the trajectory we need to provide a spacecraft template.
88    let sc_template = Spacecraft::builder()
89        .mass(Mass::from_dry_and_prop_masses(1018.0, 900.0)) // Launch masses
90        .srp(SRPData {
91            // SRP configuration is arbitrary, but we will be estimating it anyway.
92            area_m2: 3.9 * 2.7,
93            coeff_reflectivity: 0.96,
94        })
95        .orbit(Orbit::zero(MOON_J2000)) // Setting a zero orbit here because it's just a template
96        .build();
97    // Now we can build the trajectory from the BSP file.
98    // We'll arbitrarily set the tracking arc to 24 hours with a five second time step.
99    let traj_as_flown = Traj::from_bsp(
100        lro_frame,
101        MOON_J2000,
102        almanac.clone(),
103        sc_template,
104        5.seconds(),
105        Some(Epoch::from_str("2024-01-01 00:00:00 UTC")?),
106        Some(Epoch::from_str("2024-01-02 00:00:00 UTC")?),
107        Aberration::LT,
108        Some("LRO".to_string()),
109    )?;
110
111    println!("{traj_as_flown}");
112
113    // ====================== //
114    // === MODEL MATCHING === //
115    // ====================== //
116
117    // Set up the spacecraft dynamics.
118
119    // Specify that the orbital dynamics must account for the graviational pull of the Earth and the Sun.
120    // The gravity of the Moon will also be accounted for since the spaceraft in a lunar orbit.
121    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![EARTH, SUN, JUPITER_BARYCENTER]);
122
123    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
124    // We're using the GRAIL JGGRX model.
125    let mut jggrx_meta = MetaFile {
126        uri: "http://public-data.nyxspace.com/nyx/models/Luna_jggrx_1500e_sha.tab.gz".to_string(),
127        crc32: Some(0x6bcacda8), // Specifying the CRC32 avoids redownloading it if it's cached.
128    };
129    // And let's download it if we don't have it yet.
130    jggrx_meta.process(true)?;
131
132    // Build the spherical harmonics.
133    // The harmonics must be computed in the body fixed frame.
134    // We're using the long term prediction of the Moon principal axes frame.
135    let moon_pa_frame = MOON_PA_FRAME.with_orient(31008);
136    let sph_harmonics = GravityField::new(GravityFieldData::from_shadr(
137        &jggrx_meta.uri,
138        80,
139        80,
140        true,
141        almanac.frame_info(moon_pa_frame)?,
142    )?);
143
144    // Include the spherical harmonics into the orbital dynamics.
145    orbital_dyn.accel_models.push(sph_harmonics);
146
147    // We define the solar radiation pressure, using the default solar flux and accounting only
148    // for the eclipsing caused by the Earth and Moon.
149    // Note that by default, enabling the SolarPressure model will also enable the estimation of the coefficient of reflectivity.
150    let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], &almanac)?;
151
152    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
153    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
154    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
155
156    println!("{dynamics}");
157
158    // Now we can build the propagator.
159    let setup = Propagator::default_dp78(dynamics.clone());
160
161    // For reference, let's build the trajectory with Nyx's models from that LRO state.
162    let (sim_final, traj_as_sim) = setup
163        .with(*traj_as_flown.first(), almanac.clone())
164        .until_epoch_with_traj(traj_as_flown.last().epoch())?;
165
166    println!("SIM INIT:  {:x}", traj_as_flown.first());
167    println!("SIM FINAL: {sim_final:x}");
168    // Compute RIC difference between SIM and LRO ephem
169    let sim_lro_delta = sim_final
170        .orbit
171        .ric_difference(&traj_as_flown.last().orbit)?;
172    println!("{traj_as_sim}");
173    println!(
174        "SIM v LRO - RIC Position (m): {:.3}",
175        sim_lro_delta.radius_km * 1e3
176    );
177    println!(
178        "SIM v LRO - RIC Velocity (m/s): {:.3}",
179        sim_lro_delta.velocity_km_s * 1e3
180    );
181
182    traj_as_sim.ric_diff_to_parquet(
183        &traj_as_flown,
184        output_folder.join("./04_lro_sim_truth_error.parquet"),
185        ExportCfg::default(),
186    )?;
187
188    // ==================== //
189    // === OD SIMULATOR === //
190    // ==================== //
191
192    // After quite some time trying to exactly match the model, we still end up with an oscillatory difference on the order of 150 meters between the propagated state
193    // and the truth LRO state.
194
195    // Therefore, we will actually run an estimation from a dispersed LRO state.
196    // The sc_seed is the true LRO state from the BSP.
197    let sc_seed = *traj_as_flown.first();
198
199    // Load the Deep Space Network ground stations.
200    // Nyx allows you to build these at runtime but it's pretty static so we can just load them from YAML.
201    let ground_station_file: PathBuf = [
202        env!("CARGO_MANIFEST_DIR"),
203        "examples",
204        "04_lro_od",
205        "dsn-network.yaml",
206    ]
207    .iter()
208    .collect();
209
210    let devices = GroundStation::load_named(ground_station_file)?;
211
212    let mut proc_devices = devices.clone();
213
214    // Increase the noise in the devices to accept more measurements.
215    for gs in proc_devices.values_mut() {
216        if let Some(noise) = &mut gs
217            .stochastic_noises
218            .as_mut()
219            .unwrap()
220            .get_mut(&MeasurementType::Range)
221        {
222            *noise.white_noise.as_mut().unwrap() *= 3.0;
223        }
224    }
225
226    // Typical OD software requires that you specify your own tracking schedule or you'll have overlapping measurements.
227    // Nyx can build a tracking schedule for you based on the first station with access.
228    let trkconfg_yaml: PathBuf = [
229        env!("CARGO_MANIFEST_DIR"),
230        "examples",
231        "04_lro_od",
232        "tracking-cfg.yaml",
233    ]
234    .iter()
235    .collect();
236
237    let configs: BTreeMap<String, TrkConfig> = TrkConfig::load_named(trkconfg_yaml)?;
238
239    // Build the tracking arc simulation to generate a "standard measurement".
240    let mut trk = TrackingArcSim::<Spacecraft, GroundStation>::with_seed(
241        devices.clone(),
242        traj_as_flown.clone(),
243        configs,
244        123, // Set a seed for reproducibility
245    )?;
246
247    trk.build_schedule(almanac.clone())?;
248    let arc = trk.generate_measurements(almanac.clone())?;
249    // Save the simulated tracking data
250    arc.to_parquet_simple(output_folder.join("04_lro_simulated_tracking.parquet"))?;
251
252    // We'll note that in our case, we have continuous coverage of LRO when the vehicle is not behind the Moon.
253    println!("{arc}");
254
255    // Now that we have simulated measurements, we'll run the orbit determination.
256
257    // ===================== //
258    // === OD ESTIMATION === //
259    // ===================== //
260
261    let sc = SpacecraftUncertainty::builder()
262        .nominal(sc_seed)
263        .frame(LocalFrame::RIC)
264        .x_km(0.5)
265        .y_km(0.5)
266        .z_km(0.5)
267        .vx_km_s(5e-3)
268        .vy_km_s(5e-3)
269        .vz_km_s(5e-3)
270        .build();
271
272    // Build the filter initial estimate, which we will reuse in the filter.
273    let mut initial_estimate = sc.to_estimate()?;
274    initial_estimate.covar *= 3.0;
275
276    println!("== FILTER STATE ==\n{sc_seed:x}\n{initial_estimate}");
277
278    // Build the SNC in the Moon J2000 frame, specified as a velocity noise over time.
279    let process_noise = ProcessNoise3D::from_velocity_km_s(
280        &[1e-12, 1e-12, 1e-12],
281        1 * Unit::Hour,
282        10 * Unit::Minute,
283        None,
284    );
285
286    println!("{process_noise}");
287
288    // We'll set up the OD process to reject measurements whose residuals are move than 3 sigmas away from what we expect.
289    let odp = SpacecraftKalmanOD::new(
290        setup,
291        KalmanVariant::ReferenceUpdate,
292        Some(ResidRejectCrit::default()),
293        proc_devices,
294        almanac.clone(),
295    )
296    .with_process_noise(process_noise);
297
298    let od_sol = odp.process_arc(initial_estimate, &arc)?;
299
300    let final_est = od_sol.estimates.last().unwrap();
301
302    println!("{final_est}");
303
304    let ric_err = traj_as_flown
305        .at(final_est.epoch())?
306        .orbit
307        .ric_difference(&final_est.orbital_state())?;
308    println!("== RIC at end ==");
309    println!("RIC Position (m): {:.3}", ric_err.radius_km * 1e3);
310    println!("RIC Velocity (m/s): {:.3}", ric_err.velocity_km_s * 1e3);
311
312    println!(
313        "Num residuals rejected: #{}",
314        od_sol.rejected_residuals().len()
315    );
316    println!(
317        "Percentage within +/-3: {}",
318        od_sol.residual_ratio_within_threshold(3.0).unwrap()
319    );
320    println!("Ratios normal? {}", od_sol.is_normal(None).unwrap());
321
322    od_sol.to_parquet(
323        output_folder.join("04_lro_od_results.parquet"),
324        ExportCfg::default(),
325    )?;
326
327    // Create the ephemeris
328    let ephem = od_sol.to_ephemeris("LRO rebuilt".to_string());
329    let ephem_start = ephem.start_epoch().unwrap();
330    let ephem_end = ephem.end_epoch().unwrap();
331    // Check that the covariance is PSD throughout the ephemeris by interpolating it.
332    for epoch in TimeSeries::inclusive(ephem_start, ephem_end, Unit::Minute * 5) {
333        ephem
334            .covar_at(
335                epoch,
336                anise::ephemerides::ephemeris::LocalFrame::RIC,
337                &almanac,
338            )
339            .unwrap_or_else(|e| panic!("covar not PSD at {epoch}: {e}"));
340    }
341    // Export as BSP!
342    ephem
343        .write_spice_bsp(
344            -85,
345            output_folder.join("04_lro_rebuilt.bsp").to_str().unwrap(),
346            None,
347        )
348        .expect("could not built BSP");
349    let new_almanac = Almanac::default()
350        .load(output_folder.join("04_lro_rebuilt.bsp").to_str().unwrap())
351        .unwrap();
352    new_almanac.describe(None, None, None, None, None, None, None, None);
353    let (spk_start, spk_end) = new_almanac.spk_domain(-85).unwrap();
354
355    assert!((ephem_start - spk_start).abs() < Unit::Microsecond * 1);
356    assert!((ephem_end - spk_end).abs() < Unit::Microsecond * 1);
357
358    // In our case, we have the truth trajectory from NASA.
359    // So we can compute the RIC state difference between the real LRO ephem and what we've just estimated.
360    // Export the OD trajectory first.
361    let od_trajectory = od_sol.to_traj()?;
362    // Build the RIC difference.
363    od_trajectory.ric_diff_to_parquet(
364        &traj_as_flown,
365        output_folder.join("04_lro_od_truth_error.parquet"),
366        ExportCfg::default(),
367    )?;
368
369    Ok(())
370}

pub const fn signum(&self) -> i8

Returns the sign of this duration

  • 0 if the number is zero
  • 1 if the number is positive
  • -1 if the number is negative

pub fn decompose(&self) -> (i8, u64, u64, u64, u64, u64, u64, u64)

Decomposes a Duration in its sign, days, hours, minutes, seconds, ms, us, ns

pub fn subdivision(&self, unit: Unit) -> Option<Duration>

Returns the subdivision of duration in this unit, if such is available. Does not work with Week or Century.

Example
use hifitime::{Duration, TimeUnits, Unit};

let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.subdivision(Unit::Hour), Some(2.hours()));
assert_eq!(two_hours_three_min.subdivision(Unit::Minute), Some(3.minutes()));
assert_eq!(two_hours_three_min.subdivision(Unit::Second), Some(Duration::ZERO));
assert_eq!(two_hours_three_min.subdivision(Unit::Week), None);

pub fn floor(&self, duration: Duration) -> Duration

Floors this duration to the closest duration from the bottom

Example
use hifitime::{Duration, TimeUnits};

let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.floor(1.hours()), 2.hours());
assert_eq!(two_hours_three_min.floor(30.minutes()), 2.hours());
// This is zero because we floor by a duration longer than the current duration, rounding it down
assert_eq!(two_hours_three_min.floor(4.hours()), 0.hours());
assert_eq!(two_hours_three_min.floor(1.seconds()), two_hours_three_min);
assert_eq!(two_hours_three_min.floor(1.hours() + 1.minutes()), 2.hours() + 2.minutes());
assert_eq!(two_hours_three_min.floor(1.hours() + 5.minutes()), 1.hours() + 5.minutes());

pub fn ceil(&self, duration: Duration) -> Duration

Ceils this duration to the closest provided duration

This simply floors then adds the requested duration

Example
use hifitime::{Duration, TimeUnits};

let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.ceil(1.hours()), 3.hours());
assert_eq!(two_hours_three_min.ceil(30.minutes()), 2.hours() + 30.minutes());
assert_eq!(two_hours_three_min.ceil(4.hours()), 4.hours());
assert_eq!(two_hours_three_min.ceil(1.seconds()), two_hours_three_min + 1.seconds());
assert_eq!(two_hours_three_min.ceil(1.hours() + 5.minutes()), 2.hours() + 10.minutes());

pub fn round(&self, duration: Duration) -> Duration

Rounds this duration to the closest provided duration

This performs both a ceil and floor and returns the value which is the closest to current one.

Example
use hifitime::{Duration, TimeUnits};

let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.round(1.hours()), 2.hours());
assert_eq!(two_hours_three_min.round(30.minutes()), 2.hours());
assert_eq!(two_hours_three_min.round(4.hours()), 4.hours());
assert_eq!(two_hours_three_min.round(1.seconds()), two_hours_three_min);
assert_eq!(two_hours_three_min.round(1.hours() + 5.minutes()), 2.hours() + 10.minutes());

pub fn approx(&self) -> Duration

Rounds this duration to the largest units represented in this duration.

This is useful to provide an approximate human duration. Under the hood, this function uses round, so the “tipping point” of the rounding is half way to the next increment of the greatest unit. As shown below, one example is that 35 hours and 59 minutes rounds to 1 day, but 36 hours and 1 minute rounds to 2 days because 2 days is closer to 36h 1 min than 36h 1 min is to 1 day.

Example
use hifitime::{Duration, TimeUnits};

assert_eq!((2.hours() + 3.minutes()).approx(), 2.hours());
assert_eq!((24.hours() + 3.minutes()).approx(), 1.days());
assert_eq!((35.hours() + 59.minutes()).approx(), 1.days());
assert_eq!((36.hours() + 1.minutes()).approx(), 2.days());
assert_eq!((47.hours() + 3.minutes()).approx(), 2.days());
assert_eq!((49.hours() + 3.minutes()).approx(), 2.days());

pub fn min(self, other: Duration) -> Duration

use hifitime::TimeUnits;

let d0 = 20.seconds();
let d1 = 21.seconds();

assert_eq!(d0, d1.min(d0));
assert_eq!(d0, d0.min(d1));

pub fn max(self, other: Duration) -> Duration

Returns the maximum of the two durations.

use hifitime::TimeUnits;

let d0 = 20.seconds();
let d1 = 21.seconds();

assert_eq!(d1, d1.max(d0));
assert_eq!(d1, d0.max(d1));

pub const fn is_negative(&self) -> bool

Returns whether this is a negative or positive duration.