pub struct NormalizedConsistency {
pub normalized_sum: f64,
pub k: f64,
pub lower_bound: f64,
pub upper_bound: f64,
pub is_nees: bool,
}Fields
normalized_sum: f64k: f64lower_bound: f64upper_bound: f64is_nees: boolImplementations
Sourceimpl NormalizedConsistency
impl NormalizedConsistency
pub fn name(&self) -> &'static str
Sourcepub fn has_statistical_power(&self) -> bool
pub fn has_statistical_power(&self) -> bool
Returns true if there are more than 35 degrees of freedom
Sourcepub fn is_consistent(&self) -> bool
pub fn is_consistent(&self) -> bool
Returns true if the sum metric is within the lower and upper bounds
Sourcepub fn is_underconfident(&self) -> bool
pub fn is_underconfident(&self) -> bool
Returns true if the normalized sum is less than the lower bound
Sourcepub fn is_overconfident(&self) -> bool
pub fn is_overconfident(&self) -> bool
Returns true if the normalized sum is greater than the upper bound