pub struct SigmaRejection {
pub num_sigmas: f64,
}Expand description
Reject measurements if the prefit is greater than the provided sigmas deviation from the measurement noise.
Important
Some software, like ODTK, processes each measurement as a scalar. Nyx can process the measurements together. As such, if the prefit on range is bad, then the Doppler measurement with the same time stamp will also be rejected. This can lead to better convergence of the filter, and more appropriate results.
Fields
num_sigmas: f64Number of sigmas for a measurement to be considered an outlier.